﻿; ------------------------------------------------------------------------------------
; Easy Network Manager - Authentifikations System
; PB 4.5+ OS: Windows/Linux/MacOS Architecture: x86/x64
; Author: Alexander Aigner, cxAlex @ purebasic.fr/german
; ------------------------------------------------------------------------------------

EnableExplicit

Enumeration
  #ENM_Auth_Ok ; Authentifikation Ok
  #ENM_Auth_FalseUserNamePassword ; Falscher Benutzername, Passwort
  #ENM_Auth_MaximumLoginCountReached ; Zu viele Logins
  #ENM_Auth_AccountLocked ; Account gesperrt
  #ENM_Auth_InvalidToken ; Ungültiger Token
  #ENM_Auth_TokenExpired ; Token abgelaufen
  #ENM_Auth_Panic ; Panic ausgelöst, keine Logins, alle Sessions werden ungültig!
EndEnumeration

; Zugriffsrechte für die User
Enumeration
  #ENM_Auth_User
  #ENM_Auth_Operator
  #ENM_Auth_Admin
EndEnumeration

; UserName
Structure _ENM_Auth_User
  *VTable.i ; VTable
  *ENMA._ENM_Auth_Data ; Auth-System
  Permissions.i ; Zugriffsrechte
  GMutex.i ; Schutzmutex
  Name$ ; Benutzername
  Password$ ; Passwort
  Token$ ; Zugriffstoken
  TokenExpiration.i ; Verfallsdatum des Token
  TokenTimeToLife.i ; Token Verfalls zeit
  LoginCount.i ; Zugriffszähler
  MaximumLoginCount.i ; Maximale Anzahl der Zugriffe
  XData.i ; Benutzerdaten
  Locked.a ; Zugriff gesperrt
  MasterAccess.a ; Der Zugriff findet via Master-Access statt
EndStructure

Structure _ENM_Auth_Defaults
  TokenTimeToLife.i ; Wie lange ist ein Zugriffs-Token gültig?
  LoginCount.i ; Wie viele gleichzeutige Logins pro User sind zulässig?
EndStructure

Structure _ENM_Auth_Data
  *VTable.i ; VTable
  *ENM._ENM_Data ; ENM Daten
  Map Users._ENM_Auth_User(2*#ENM_Sys_ServerClientSlots) ; Benutzernamen
  Map *AuthTokens(2*#ENM_Sys_ServerClientSlots) ; AuthToken
  ; Defaults
  Defaults._ENM_Auth_Defaults
  ; Schutzmutexen
  GUserMutex.i
  GAuthMutex.i
  ; Sofort alles abschalten!
  Panic.a
  ; Master - Zugriff
  MasterAccessPassword$ ; Master Password
  MasterAccessToken$ ; Master Token
  MasterAccessTokenExpiration.i ; Ablauf des Mastertoken
EndStructure

; System - Interface
Interface ENM_AuthenticationSystem
  AddUser(UserName$, Password$, Permissions = #ENM_Auth_User) ; Benutzer hinzufügen
  Login(UserName$, PasswordHash$, *LoginResult.Integer = #Null) ; Einloggen
  LoginPerToken(Token$, *LoginResult.Integer = #Null) ; Einloggen per Token
  Exporter(ExportCB) ; Alles Exportieren
  Importer(ImportCB) ; Alles Importieren
  Panic() ; Panic, keine Loggins mehr möglich, alle Sessions werden ungültig
  Free() ; System freigeben
EndInterface

; User - Interface
Interface ENM_AuthenticationSystem_User
  IsSessionValid() ; Ist der User immer noch korrekt eingeloggt?
  GenerateAccessToken.s() ; Access-Token erstellen
  DeleteAccessToken() ; Access-Token löschen
  GetData() ; Benutzerdaten laden
  SetData(XData.i) ; Benutzerdaten setzen
  GetName.s() ; Namen laden
  GetPermissions() ; Zugriffsrechte
  Logout(KeepToken = #True) ; User ausloggen
  ; Admin - Funktionen, benötigen Master - Access!
  GetMasterAccess(MasterPasswordHash$) ; Master Zugriff anfordern
  IsMaster() ; Haben wir Master Access?
  LockUser(UserName$, Lock) ; Sperrt oder Entsperrt einen User
  DeleteUser(UserName$) ; Benutzer löschen
  SetPermissions(UserName$, Permissions) ; setzt die Permissions für einen User
  SetMaximumLoginCount(UserName$, Count) ; setzt den maximalen Login.Count für einen User
  SetTokenTimeToLife(UserName$, TokenTimeToLife) ; Setzt die Token-TTL für einen User.
  ForceLogin(UserName$) ; Einloggen in einen User ohne irgendwelche Abfragen
EndInterface

; Passwort Hashen
Procedure.s ENM_Auth_HashPassword(Password$)
  ProcedureReturn "#hash:"+MD5Fingerprint(@Password$, StringByteLength(Password$))
EndProcedure

; Export - Aufruf
Prototype ENM_Auth_ExportImportCallback(ENMA.ENM_AuthenticationSystem, UserName$, PasswordHash$, Permission, XData, Token$, TokenExpiration, Locked, MaximumLoginCount, TokenTimeToLife)
; Import Aufruf
Prototype ENM_Auth_Importer(ENMA.ENM_AuthenticationSystem, Importer.ENM_Auth_ExportImportCallback)

; Zufälligen Token erstellen
Procedure.s __ENM_Auth_GenerateToken()
  Protected *Buffer = AllocateMemory(172)
  Protected Result$, length
  
  If OpenCryptRandom()
    CryptRandomData(*Buffer, 64)
    CloseCryptRandom()
    
    length = Base64Encoder(*Buffer, 64, *Buffer+64, 128)
    Result$ = PeekS(*Buffer+64, length)
    FreeMemory(*Buffer)
  EndIf
  
  ProcedureReturn Result$
EndProcedure

; Daten importieren
; diese Procedure wird innerhalb der gelockten Mutexen aufgerufen, kein Lock nötig!
Procedure __ENM_Auth_Importer(ENMA.ENM_AuthenticationSystem, UserName$, Password$, Permission, XData, Token$, TokenExpiration, Locked, MaximumLoginCount, TokenTimeToLife)
  Protected *ENMA._ENM_Auth_Data = ENMA
  Protected *User._ENM_Auth_User, User$
  With *ENMA
    ; User wird als LowerCase gespeichert...
    User$ = LCase(UserName$)
    ; User anlegen falls nicht vorhanden
    *User = FindMapElement(\Users(), User$)
    If Not *User
      *User = AddMapElement(\Users(), User$)
    EndIf
    
    If *User ; Benutzer wurde angelegt, Daten importieren
      *User\Name$ = UserName$
      *User\Password$ = Password$
      *User\Permissions = Permission
      *User\XData = XData
      *User\Token$ = Token$
      *User\TokenExpiration = TokenExpiration
      *User\MaximumLoginCount = MaximumLoginCount
      *User\TokenTimeToLife = TokenTimeToLife
      *User\Locked = Locked
      ; Token anlegen
      If *User\Token$
        AddMapElement(\AuthTokens(), *User\Token$) : \AuthTokens() = *User 
      EndIf
    EndIf
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
; System - Funktionen
; ------------------------------------------------------------------------------------

; Neuen Authentifikator anlegen
Procedure _ENM_Auth_New(*ENM._ENM_Data, MasterPassword$, MaximumLoginCount = 1, TokenTimeToLife = 604800)
  Protected *ENMA._ENM_Auth_Data = AllocateMemory(SizeOf(_ENM_Auth_Data))
  InitializeStructure(*ENMA, _ENM_Auth_Data)
  With *ENMA
    ; VTable
    \VTable = ?_ENM_Auth_VTable
    ; ENM speichern
    \ENM = *ENM
    ; Schutzmutexen anlegen
    \GUserMutex = CreateMutex()
    \GAuthMutex  = CreateMutex()
    ; Defaultwerte setzen:
    \Defaults\LoginCount = MaximumLoginCount ; ein Login
    \Defaults\TokenTimeToLife = TokenTimeToLife ; eine Woche
    ; Passwort für den Master - Modus
    If (Left(MasterPassword$,6) = "#hash:") And (Len(MasterPassword$) = 38) ; Passwort ist schon gehasht
      \MasterAccessPassword$ = MasterPassword$
    Else  
      \MasterAccessPassword$ = ENM_Auth_HashPassword(MasterPassword$)
    EndIf
    
  EndWith
  ProcedureReturn *ENMA
EndProcedure

; Authentication System freigeben. Async nicht nötig
Procedure _ENM_Auth_Free(*ENMA._ENM_Auth_Data)
  Protected *User._ENM_Auth_User
  With *ENMA
    \Panic = #True ; Ich liebe Panic...
    
    LockMutex(\GUserMutex)
    While MapSize(\Users())
      If Not NextMapElement(\Users())
        ResetMap(\Users())
        NextMapElement(\Users())
      EndIf
      *User = \Users()
      ; User rausschmeißen falls nötig...
      If *User\LoginCount
        *User\Locked = #True
        UnlockMutex(\GUserMutex)
        While *User\LoginCount
          Delay(1)
        Wend
        LockMutex(\GUserMutex)
      EndIf
      ; Mutex freigeben
      LockMutex(*User\GMutex) : UnlockMutex(*User\GMutex) ; Letzter Lock-Zyklus
      FreeMutex(*User\GMutex)
      ; User freigeben
      DeleteMapElement(\Users(), LCase(*User\Name$))
    Wend
    UnlockMutex(\GUserMutex)
    
    ; Mutexen freigeben
    FreeMutex(\GUserMutex)
    FreeMutex(\GAuthMutex)
    
    ClearStructure(*ENMA, _ENM_Auth_Data)
    FreeMemory(*ENMA)
  EndWith
EndProcedure

CompilerIf #ENM_Sys_AsyncFree
  ; Freigabe für den Async-Manager
  Procedure _ENM_Auth_Free_Async(*ENMA._ENM_Auth_Data)
    With *ENMA
      If ENM
        JQ_Add(\ENM\GBCollector, @_ENM_Auth_Free(), *ENMA)
      Else
        _ENM_Auth_Free(*ENMA)
      EndIf
    EndWith
  EndProcedure
CompilerEndIf

; Authentication System exportieren
Procedure _ENM_Auth_Export(*ENMA._ENM_Auth_Data, Exporter.ENM_Auth_ExportImportCallback)
  Protected *User._ENM_Auth_User
  With *ENMA
    ; KEIN! Panic-Lock, der Export muss immer noch funktionierten, sonst könnten Daten verloren gehen
    
    LockMutex(\GUserMutex)
    ForEach \Users()
      *User = \Users()
      ; Export - CB für jeden Eintrag aufrufen
      Exporter(*ENMA, *User\Name$, *User\Password$, *User\Permissions, *User\XData, *User\Token$, *User\TokenExpiration, *User\Locked, *User\MaximumLoginCount, *User\TokenTimeToLife)
    Next
    UnlockMutex(\GUserMutex)
    
    ProcedureReturn #Null
  EndWith
EndProcedure

; Importer
Procedure _ENM_Auth_Import(*ENMA._ENM_Auth_Data, Importer.ENM_Auth_Importer)
  With *ENMA
    If \Panic ; Panic, nichts geht mehr
      ProcedureReturn #False
    EndIf
    
    LockMutex(\GAuthMutex)
    LockMutex(\GUserMutex)
    Importer(*ENMA, @__ENM_Auth_Importer()) ; Importer aufrufen
    UnlockMutex(\GUserMutex)
    UnlockMutex(\GAuthMutex)
    
    ProcedureReturn #True
  EndWith
EndProcedure

; Neuen User anlegen
Procedure _ENM_Auth_AddUser(*ENMA._ENM_Auth_Data, User$, Password$, Permissions = #ENM_Auth_User)
  Protected UserName$, *User._ENM_Auth_User
  
  With *ENMA
    If \Panic ; Panic, nichts geht mehr...
      ProcedureReturn #Null
    EndIf
    
    ; Username beleibt, eingetragen wird in Lower-Case
    UserName$ = User$
    User$ = LCase(User$)
    
    ; Gibt es den User?
    LockMutex(\GUserMutex)
    If Not FindMapElement(\Users(), User$)
      *User = AddMapElement(\Users(), User$)
    EndIf
    UnlockMutex(\GUserMutex)
    
    ; Daten eintragen:
    If *User
      *User\VTable = ?_ENM_AuthUser_VTable ; VTable
      *User\GMutex = CreateMutex() ; Schutzmutex
      *User\ENMA = *ENMA ; *ENMA Zugriff
      *User\Name$ = UserName$ ; Benutzername
      If (Left(Password$,6) = "#hash:") And (Len(Password$) = 38) ; Passwort ist schon gehasht
        *User\Password$ = Password$
      Else  
        *User\Password$ = ENM_Auth_HashPassword(Password$) ; Passwort-Hash
      EndIf
      *User\MaximumLoginCount = \Defaults\LoginCount ; Maximale Zugriffszahl
      *User\TokenTimeToLife = \Defaults\TokenTimeToLife ; Lebenszeit eines Zugriffstokens
      *User\Permissions = Permissions ; Berechtigungsstufe
      
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; Einloggen
Procedure _ENM_Auth_Login(*ENMA._ENM_Auth_Data, User$, PasswordHash$, *LoginResult.Integer = #Null)
  Protected *User._ENM_Auth_User
  Protected RtDummy
  ; Der Programmierer will den Status nicht wissen...
  If Not *LoginResult
    *LoginResult = @RtDummy
  EndIf
  
  ; Loginstatus setzen
  *LoginResult\i = #ENM_Auth_FalseUserNamePassword
  With *ENMA
    If \Panic ; Panic, nichts geht mehr
      *LoginResult\i = #ENM_Auth_Panic
      ProcedureReturn #Null
    EndIf
    ; Username ist immer Lower-Case
    User$ = LCase(User$)
    ; User laden falls vorhanden
    LockMutex(\GUserMutex)
    *User = FindMapElement(\Users(), User$)
    UnlockMutex(\GUserMutex)
    
    If *User ; Git es den User?
      If *User\Password$ = PasswordHash$ ; Passwort Ok?
        If Not *User\Locked
          If *User\MaximumLoginCount And (*User\LoginCount<*User\MaximumLoginCount) ; Sind wir noch nicht zu oft eingeloggt?
            *LoginResult\i = #ENM_Auth_Ok ; Alles Ok...
            LockMutex(*User\GMutex)
            *User\LoginCount + 1 ; Loginzähler erhöhen
            UnlockMutex(*User\GMutex)
          Else ; Zu viele Logins...
            *User = #Null
            *LoginResult\i = #ENM_Auth_MaximumLoginCountReached
          EndIf
        Else ; Account gesperrt!
          *User = #Null
          *LoginResult\i = #ENM_Auth_AccountLocked
        EndIf
      Else
        *User = #Null
      EndIf
    EndIf
    
    ; Benutzer zurückgeben
    ProcedureReturn *User
  EndWith
EndProcedure

; Einloggen per Token
Procedure _ENM_Auth_LoginPerToken(*ENMA._ENM_Auth_Data, Token$, *LoginResult.Integer = #Null)
  Protected *User._ENM_Auth_User
  Protected RtDummy
  ; Der Programmierer will den Status nicht wissen...
  If Not *LoginResult
    *LoginResult = @RtDummy
  EndIf
  *LoginResult\i = #ENM_Auth_InvalidToken
  
  With *ENMA
    If \Panic ; Panic, nichts geht mehr
      *LoginResult\i = #ENM_Auth_Panic
      ProcedureReturn #Null
    EndIf
    
    ; Token suchen
    LockMutex(\GAuthMutex)
    *User = FindMapElement(\AuthTokens(), Token$)
    If *User ; Pointer.Pointer ;)
      *User = PeekI(*User)
    EndIf
    UnlockMutex(\GAuthMutex)
    
    If *User ; Ah, der Token ist ok!
      If Not *User\Locked
        If Date() <= *User\TokenExpiration ; Token noch gültig?
          If *User\MaximumLoginCount And (*User\LoginCount<>*User\MaximumLoginCount) ; Sind wir noch nicht zu oft eingeloggt?
            *LoginResult\i = #ENM_Auth_Ok ; Alles Ok...
            LockMutex(*User\GMutex)
            *User\LoginCount + 1 ; Loginzähler erhöhen
            UnlockMutex(*User\GMutex)
          Else ; Zu viele Logins...
            *User = #Null
            *LoginResult\i = #ENM_Auth_MaximumLoginCountReached
          EndIf
        Else ; Token abgelaufen
          *User = #Null
          *LoginResult\i = #ENM_Auth_TokenExpired
          ; Token löschen:
          LockMutex(\GAuthMutex)
          DeleteMapElement(\AuthTokens(), Token$)
          UnlockMutex(\GAuthMutex)
        EndIf
      Else ; Account gesperrt
        *User = #Null
        *LoginResult\i = #ENM_Auth_AccountLocked
      EndIf
    EndIf
    
    ; Benutzer zurückgeben
    ProcedureReturn *User
  EndWith
EndProcedure

; Panic auslösen
Procedure _ENM_Auth_Panic(*ENMA._ENM_Auth_Data)
  With *ENMA
    \Panic = #True
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
; User - Funktionen
; ------------------------------------------------------------------------------------

; Session validieren
Procedure _ENM_Auth_User_IsSessionValid(*User._ENM_Auth_User)
  With *User
    ; Nicht gesperrt und keine Panic
    ProcedureReturn ((Not (\Locked Or \ENMA\Panic)) Or #Null)
  EndWith
EndProcedure

; Token löschen
Procedure _ENM_Auth_User_DeleteAccessToken(*User._ENM_Auth_User)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #Null
    EndIf
    
    ; Token löschen...
    LockMutex(\GMutex)
    If \Token$
      LockMutex(\ENMA\GAuthMutex)
      DeleteMapElement(\ENMA\AuthTokens(), \Token$)
      UnlockMutex(\ENMA\GAuthMutex)
      ; Token löschen
      \Token$ = #NULL$
      \TokenExpiration = #Null
    EndIf
    UnlockMutex(\GMutex)
  EndWith
EndProcedure

; User ausloggen
Procedure _ENM_Auth_User_Logout(*User._ENM_Auth_User, KeepToken = #True)
  With *User
    ; Token löschen falls nötig
    If Not KeepToken
      _ENM_Auth_User_DeleteAccessToken(*User)
    EndIf
    
    ; Benutzer austragen
    LockMutex(*User\GMutex)
    *User\LoginCount - 1 ; Loginzähler erniedrigen
    UnlockMutex(*User\GMutex)
  EndWith
EndProcedure

; Zugriffstoken erstellen falls nötig
Procedure.s _ENM_Auth_User_GenerateAccessToken(*User._ENM_Auth_User)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #NULL$
    EndIf
    
    LockMutex(\GMutex)
    If Date() > \TokenExpiration
      ; Vorherigen Token austragen...
      If \Token$
         _ENM_Auth_User_DeleteAccessToken(*User)
      EndIf
    
      ; Token erstellen
      \Token$ = __ENM_Auth_GenerateToken()
      \TokenExpiration = Date() + \TokenTimeToLife
    
      ; Token eintragen
      LockMutex(\ENMA\GAuthMutex)
      AddMapElement(\ENMA\AuthTokens(), \Token$) : \ENMA\AuthTokens() = *User 
      UnlockMutex(\ENMA\GAuthMutex)
    EndIf
    UnlockMutex(\GMutex)
    
    ProcedureReturn \Token$ ; Token zurückgeben
  EndWith
EndProcedure

Procedure.s _ENM_Auth_User_GetName(*User._ENM_Auth_User)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #NULL$  
    EndIf
    
    ; Berechtigungen zurückgeben
    ProcedureReturn \Name$
  EndWith
EndProcedure

; Berechtigungen laden
Procedure _ENM_Auth_User_GetPermissions(*User._ENM_Auth_User)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #Null  
    EndIf
    
    ; Berechtigungen zurückgeben
    ProcedureReturn \Permissions
  EndWith
EndProcedure

; Berechtigungen laden
Procedure _ENM_Auth_User_GetData(*User._ENM_Auth_User)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #Null  
    EndIf
    
    ; Daten zurückgeben
    ProcedureReturn \XData
  EndWith
EndProcedure

; Berechtigungen laden
Procedure _ENM_Auth_User_SetData(*User._ENM_Auth_User, XData)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #Null  
    EndIf
    
    ; Daten setzen
    \XData = XData
    ProcedureReturn #True
  EndWith
EndProcedure

; Master - Zugriff anfordern. Zugriffsberechtigungen > USER nötig
Procedure _ENM_Auth_User_GetMasterAccess(*User._ENM_Auth_User, MasterPasswordHash$)
  With *User
    If Not _ENM_Auth_User_IsSessionValid(*User)
      ProcedureReturn #Null  
    EndIf
    
    LockMutex(\GMutex)
    ; Min. Operator ist nötig um Admin-Zugriff zu erlangen
    If (\Permissions > #ENM_Auth_User) And (\ENMA\MasterAccessPassword$ = MasterPasswordHash$)
      \MasterAccess = #True ; Wir haben Master - Zugriff!
    Else ; Wir haben keinen Master - Zugriff
      \MasterAccess = #False
    EndIf
    UnlockMutex(\GMutex)
    
    ProcedureReturn \MasterAccess
  EndWith
EndProcedure

; Is Master?
Procedure _ENM_Auth_User_IsMaster(*User._ENM_Auth_User)
  With *User
    ; Panic prüfung ist in der Valid-Routine...    
    ProcedureReturn ((_ENM_Auth_User_IsSessionValid(*User) And \MasterAccess) Or #Null)
  EndWith
EndProcedure

; User Locken
Procedure _ENM_Auth_User_LockUser(*User._ENM_Auth_User, User$, Lock)
  Protected *User2._ENM_Auth_User
  
  With *User
    ; Panic, Lock und Masterfrage
    If Not _ENM_Auth_User_IsMaster(*User)
      ProcedureReturn #Null
    EndIf
    ; User wird immer in Lower-Case gespeichert...
    User$ = LCase(User$)
    
    ; User laden Falls vorhanden
    LockMutex(\ENMA\GUserMutex)
    *User2 = FindMapElement(\ENMA\Users(), User$)
    UnlockMutex(\ENMA\GUserMutex)
    
    ; Gibt es den User?
    If *User2 = *User ; Selbstlock nicht möglich!
      *User2 = #Null
    EndIf
    
    If *User2
      LockMutex(*User2\GMutex)
      *User2\Locked = Lock ; User Locken
      UnlockMutex(*User2\GMutex)
      
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; User löschen
Procedure _ENM_Auth_User_DeleteUser(*User._ENM_Auth_User, User$)
  Protected *User2._ENM_Auth_User
  
  With *User
    ; Panic, Lock und Masterfrage
    If Not _ENM_Auth_User_IsMaster(*User)
      ProcedureReturn #Null
    EndIf
    ; User wird immer in Lower-Case gespeichert...
    User$ = LCase(User$)
    
    ; User laden Falls vorhanden
    LockMutex(\ENMA\GUserMutex)
    *User2 = FindMapElement(\ENMA\Users(), User$)
    UnlockMutex(\ENMA\GUserMutex)
    
    ; Gibt es den User?
    If *User2 = *User ; Selbstlöschung nicht möglich!
      *User2 = #Null
    EndIf
    
    If *User2
      ; User rauswerfen, warten bis Logout...
      LockMutex(*User2\GMutex)
      *User2\Locked = #True ; User Locken
      If *User2\LoginCount ; Da ist noch jemand eingeloggt...
        UnlockMutex(*User2\GMutex)
        While *User2\LoginCount
          Delay(1)
        Wend
        LockMutex(*User2\GMutex)
      EndIf
      UnlockMutex(*User2\GMutex)
      
      ; GMutex freigeben
      LockMutex(*User2\GMutex) : UnlockMutex(*User2\GMutex) ; Letzter Lock-Zyklus...
      FreeMutex(*User2\GMutex)
      
      ; Token löschen falls vorhanden...
      If *User\Token$
        LockMutex(\ENMA\GAuthMutex)
        DeleteMapElement(\ENMA\AuthTokens(), *User\Token$)
        UnlockMutex(\ENMA\GAuthMutex)
      EndIf
      
      ; User löschen
      LockMutex(\ENMA\GUserMutex)
      DeleteMapElement(\ENMA\Users(), User$)
      UnlockMutex(\ENMA\GUserMutex)
      
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; Std - Setter für Master - Funktionen
Macro _ENM_Auth_User_MasterSetX(_Name)
  Procedure _ENM_Auth_User_Set#_Name(*User._ENM_Auth_User, User$, _Name)
    Protected *User2._ENM_Auth_User
    ; Panic, Lock und Masterfrage
    If Not _ENM_Auth_User_IsMaster(*User)
      ProcedureReturn #Null
    EndIf
    ; User wird immer in Lower-Case gespeichert...
    User$ = LCase(User$)
    
    ; User laden Falls vorhanden
    LockMutex(*User\ENMA\GUserMutex)
    *User2 = FindMapElement(*User\ENMA\Users(), User$)
    UnlockMutex(*User\ENMA\GUserMutex)
    ; Gibt es den User?
    If *User2
      ; Value setzen...
      LockMutex(*User2\GMutex)
      *User2\_Name = _Name
      UnlockMutex(*User2\GMutex)
      ; Alles Ok...
      ProcedureReturn #True
    EndIf
  EndProcedure
EndMacro

; Setter erzeugen für Permissions, MaximumLoginCount, TokenTimeToLife
_ENM_Auth_User_MasterSetX(Permissions)
_ENM_Auth_User_MasterSetX(MaximumLoginCount)
_ENM_Auth_User_MasterSetX(TokenTimeToLife)

; Force Login in einen User
Procedure _ENM_Auth_User_ForceLogin(*User._ENM_Auth_User, User$)
  Protected *User2._ENM_Auth_User
  
  With *User
    ; Panic, Lock und Masterfrage
    If Not _ENM_Auth_User_IsMaster(*User)
      ProcedureReturn #Null
    EndIf
    ; User wird immer in Lower-Case gespeichert...
    User$ = LCase(User$)
    
    ; User laden Falls vorhanden
    LockMutex(\ENMA\GUserMutex)
    *User2 = FindMapElement(\ENMA\Users(), User$)
    UnlockMutex(\ENMA\GUserMutex)
    ; Gibt es den User?
    If *User2 = *User ; Doppellogin unsinnig!
      *User2 = #Null
    ElseIf *User2 ; Einloggen, egal was ist...
      LockMutex(*User2\GMutex)
      *User2\LoginCount + 1 ; Logincounter erhöhen
      UnlockMutex(*User2\GMutex)
    EndIf
    
    ; User zurückgeben
    ProcedureReturn *User2
  EndWith
EndProcedure

; VTable
DataSection
  ; System VTable
  _ENM_Auth_VTable:
  Data.i @_ENM_Auth_AddUser()
  Data.i @_ENM_Auth_Login()
  Data.i @_ENM_Auth_LoginPerToken()
  Data.i @_ENM_Auth_Export()
  Data.i @_ENM_Auth_Import()
  Data.i @_ENM_Auth_Panic()
  CompilerIf #ENM_Sys_AsyncFree
    Data.i @_ENM_Auth_Free_Async()  
  CompilerElse
    Data.i @_ENM_Auth_Free()
  CompilerEndIf
    ; User VTable
  _ENM_AuthUser_VTable:
  Data.i @_ENM_Auth_User_IsSessionValid()
  Data.i @_ENM_Auth_User_GenerateAccessToken()
  Data.i @_ENM_Auth_User_DeleteAccessToken()
  Data.i @_ENM_Auth_User_GetData()
  Data.i @_ENM_Auth_User_SetData()
  Data.i @_ENM_Auth_User_GetName()
  Data.i @_ENM_Auth_User_GetPermissions()
  Data.i @_ENM_Auth_User_Logout()
  Data.i @_ENM_Auth_User_GetMasterAccess()
  Data.i @_ENM_Auth_User_IsMaster()
  Data.i @_ENM_Auth_User_LockUser()
  Data.i @_ENM_Auth_User_DeleteUser()
  Data.i @_ENM_Auth_User_SetPermissions()
  Data.i @_ENM_Auth_User_SetMaximumLoginCount()
  Data.i @_ENM_Auth_User_SetTokenTimeToLife()
  Data.i @_ENM_Auth_User_ForceLogin()
EndDataSection

; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 479
; FirstLine = 455
; Folding = -----
; EnableThread
; EnableXP
; CPU = 1