﻿; ------------------------------------------------------------------------------------
; Dynamischer Cache für Temporärdaten
; (c) Alexander Aigner
; PB 4.5 + / x86 + x64 / Ascii + Unicode
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
;- Internal
; ------------------------------------------------------------------------------------

Interface ENM_Cache
  Add(Name$, *Mem, MemSize, MaxAge = #Null, CopyMem = #True)
  Get(Name$)
  Flush()
  Free()
  GetSize()
  CleanUp()
  Config(MaximumSize = #Null, MinimumEntrySize = 1*1024, CompressTimeLimit = 60, CompressLevel = 9)
EndInterface

Interface ENM_CacheEntry
  GetMem()
  GetMemSize()
  Release()
  Delete()
EndInterface

; EnableExplicit

; Cache Eintrag
Structure _Cache_Entry
  *Mem ; Speicher
  MemSize.i ; Speichergröße
  MaxAge.i ; Maximalalter in Sekunden
  LastAccess.i ; Letzter Zugriff
  AccessCount.i ; Zugriffszähler
  AcessMutex.i ; Zugriffsmutex
  locked.i ; Eintrag gesperrt
  deleting.a ; Eintrag wird gelöscht
  compressed.i ; Eintrag ist komprimiert
EndStructure

; Cache Daten
Structure _Cache_Data
  *VTable ; VTable für Interface
  *ENM._ENM_Data ; ENM - Zugriffsdaten
  Map CMap._Cache_Entry()
  MaxSize.i ; Maximalgröße Cache
  CurrentSize.i ; aktuelle Größe
  AccessMutex.i ; ZugriffsMutex
  CompressTimeLimit.i ; Ab wann soll komprimiert werden?
  CompressLevel.i ; wie stark soll komprimiert werden?
  MinimumEntrySize.i ; Minimalgröße eines Eintrags
EndStructure 

Structure _Cache_Entry_Data
  *VTable
  *Entry._Cache_Entry
  *Cache._Cache_Data
EndStructure

; Eintrag hat das Alterslimit überschritten
Macro _ENM_Cache_EntryisTooOld(_Entry)
  (_Entry#\MaxAge And (Date()> = _Entry#\MaxAge))
EndMacro

; aktuellen Entry löschen. Wird immer innerhalb der Accessmutex
; aufgerufen
Procedure _ENM_Cache_DeleteEntry(*Cache._Cache_Data, *CEntry._Cache_Entry, SafeDelete = #True)
  With *Cache
    LockMutex(*CEntry\AcessMutex)
    If Not *CEntry\locked Or (*CEntry\deleting And SafeDelete) ; Eintrag nicht gesperrt und wir auch noch nicht gelöscht
      \CurrentSize-*CEntry\MemSize ; Speichergröße abziehen
      ; Alles freigeben
      *CEntry\locked = #True
      UnlockMutex(*CEntry\AcessMutex)
      
      FreeMemory(*CEntry\Mem)
      FreeMutex(*CEntry\AcessMutex)
      DeleteMapElement(\CMap(), MapElementKey(*CEntry))
    Else
      UnlockMutex(*CEntry\AcessMutex)
    EndIf
  EndWith
EndProcedure

; Räumt so lange alte Einträge aus bis genug Platz ist
Procedure _ENM_Cache_ClearOld(*Cache._Cache_Data, SizeToAdd)
  Protected *CEntry._Cache_Entry, *ccEntry._Cache_Entry
  With *Cache
    If (SizeToAdd<= *Cache\MaxSize) And *Cache\MaxSize ; Passt das überhaupt rein?
      
      While (\MaxSize-\CurrentSize)<SizeToAdd ; Muss was freigeben werden?
        *ccEntry = #Null
        LockMutex(\AccessMutex) ; die Liste muss komplett durgegangen werden, kein Jumpen!
        ForEach \CMap() ; Alle abklappern
          *CEntry = \CMap() ; Eintrag holen
          
          ; Alter prüfen
          If _ENM_Cache_EntryisTooOld(*CEntry) ; Altersgrenze erreicht
            _ENM_Cache_DeleteEntry(*Cache, *CEntry)
            If (\MaxSize-\CurrentSize)>=SizeToAdd ; Es muss nichts mehr freigegeben werden
              *ccEntry = #Null ; Muss nicht mehr gelöscht werden
              Break
            EndIf
            Continue
          EndIf
          
          ; Eintrag auf dem zuletzt zugegriffen wurde der nicht gelocked ist wird gelöscht.
          ; bei gleichstand wird der eintrag auf den weniger oft zugegriffen wurde gelöscht
          If (Not *CEntry\locked) And ((Not *ccEntry) Or (*CEntry\LastAccess<*ccEntry\LastAccess) Or ((*CEntry\LastAccess=*ccEntry\LastAccess) And (*CEntry\AccessCount<*ccEntry\AccessCount)))
            *ccEntry = \CMap() ; ältester unbenutzer Eintrag
          EndIf
        Next
        
        ; ältestes Löschen
        If *ccEntry 
          _ENM_Cache_DeleteEntry(*Cache, *ccEntry)
        Else ; Keines konnte ausgewählt werden: break
          Break
        EndIf
        UnlockMutex(\AccessMutex)
      Wend
      
      ; Konnte genug freigegeben werden?
      ProcedureReturn (((\MaxSize-\CurrentSize)>=SizeToAdd) Or #Null)
    Else
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

Declare _Cache_Add(*Cache._Cache_Data, Name$, *Mem, MemSize, MaxAge = #Null, CopyMem = #True)

; ------------------------------------------------------------------------------------
;- Public
; ------------------------------------------------------------------------------------

; Cache erzeugen
; MaximumSize: Maximale Größe des Cache in Bytes, für Unbegrenzt auf #NULL lassen
; AgeTolerance: AltersToleranz bei der beim Speicherfreiräumen noch nach schlechteren Daten gesucht wird
; AccessTolerance: ZugriffssToleranz bei der beim Speicherfreiräumen noch nach schlechteren Daten gesucht wird
Procedure _ENM_Cache_New(*ENM._ENM_Data, MaximumSize = #Null)
  Protected *Cache._Cache_Data = AllocateMemory(SizeOf(_Cache_Data))
  InitializeStructure(*Cache, _Cache_Data)
  If MaximumSize < #Null
    MaximumSize = #Null
  EndIf
  With *Cache
    \ENM = *ENM
    \CompressTimeLimit = 60 ; nach 1. Minute komprimieren
    \CompressLevel = 9 ; Kompression Level
    \MinimumEntrySize = 1*1024 ; 1 KByte Mindestgröße 
    \VTable = ?_ENM_Cache_VTable
    \MaxSize = MaximumSize
    \AccessMutex = CreateMutex()
    ProcedureReturn *Cache
  EndWith
EndProcedure

; Neu Konfigurieren, Parameter siehe _Cache_New()
Procedure _ENM_Cache_Config(*Cache._Cache_Data, MaximumSize = #Null, MinimumEntrySize = 1*1024, CompressTimeLimit = 60, CompressLevel = 9)
  With *Cache
    
    ; Sinnhaftigkeit prüfen
    If MaximumSize < #Null
      MaximumSize = #Null
    EndIf
    If CompressTimeLimit < #Null
      CompressTimeLimit = #Null
    EndIf
    If CompressLevel < #Null
      CompressLevel = #Null
    EndIf
    If CompressLevel > 9
      CompressLevel = 9
    EndIf
    If MinimumEntrySize > #Null
      MinimumEntrySize = #Null
    EndIf
    
    \MaxSize = MaximumSize
    \CompressLevel = CompressLevel
    \CompressTimeLimit = CompressTimeLimit
    \MinimumEntrySize = MinimumEntrySize
  EndWith
EndProcedure

; Löscht zu alte Daten aus dem Cache
Procedure _ENM_Cache_CleanUp(*Cache._Cache_Data)
  Protected *CEntry._Cache_Entry, *Mem
  Protected packed.zlib_Info
  
  With *Cache
    LockMutex(\AccessMutex)
    ForEach \CMap()
      *CEntry = \CMap()
      If _ENM_Cache_EntryisTooOld(*CEntry) ; Altersgrenze erreicht
        _ENM_Cache_DeleteEntry(*Cache, *CEntry)
      ElseIf (\CompressTimeLimit And \CompressLevel) ; es git ein CompressTimeLimit
        
        If (Not (*CEntry\locked Or *CEntry\deleting)) ; kleiner Vortest um weniger Mutexlocks zu erzielen
          LockMutex(*CEntry\AcessMutex)
          
          ; Eintrag nicht verwendet und soll komprimiert werden
          If (Not (*CEntry\locked Or *CEntry\deleting)) And ((Date()-*CEntry\LastAccess) >= \CompressTimeLimit)
            ; Speicher packen
            *Mem = zLib_compress(*CEntry\Mem, *CEntry\MemSize,  \CompressLevel, @packed)
            If packed\DataOk ; Packen Ok
              *CEntry\compressed = packed\MemSize ; Compressed setzen
              FreeMemory(*CEntry\Mem) ; alten Speicher löschen
              *CEntry\Mem = *Mem ; Komprimierten Speicher Eintragen
            EndIf
          EndIf
          
          UnlockMutex(*CEntry\AcessMutex) 
        EndIf
      EndIf
    Next
    UnlockMutex(\AccessMutex)
  EndWith
EndProcedure

; Manuelles löschen des gesamten Cache
Procedure _ENM_Cache_Flush(*Cache._Cache_Data)
  With *Cache
    LockMutex(\AccessMutex)
    ForEach \CMap()
      _ENM_Cache_DeleteEntry(*Cache, \CMap())
    Next
    UnlockMutex(\AccessMutex)
  EndWith
EndProcedure

; Cache freigeben
Procedure _ENM_Cache_Free(*Cache._Cache_Data)
  With *Cache
    _ENM_Cache_Flush(*Cache)
    FreeMutex(\AccessMutex)
    ClearStructure(*Cache, _Cache_Data)
    FreeMemory(*Cache) ; Cache selbst freigeben
  EndWith
EndProcedure

CompilerIf #ENM_Sys_AsyncFree
  ; Freigabeprocedure für den Async - Manager
  Procedure _ENM_Cache_Free_Async(*Cache._Cache_Data)
    With *Cache
      If ENM
        JQ_Add(\ENM\GBCollector, @_ENM_Cache_Free(), *Cache)
      Else
        _ENM_Cache_Free(*Cache)
      EndIf
    EndWith
  EndProcedure
CompilerEndIf

; aktuelle Speichergröße des Cache
Procedure _ENM_Cache_GetSize(*Cache._Cache_Data)
  ProcedureReturn *Cache\CurrentSize
EndProcedure

; Nach gespeicherten Daten suchen
; Name$: Name des Cache Eintrag
Procedure _ENM_Cache_Get(*Cache._Cache_Data, Name$)
  Protected *CEntry._Cache_Entry, *CIEntry._Cache_Entry_Data, *Mem
  Name$ = LCase(Name$)
  
  With *Cache
    
    LockMutex(\AccessMutex)
    
    If FindMapElement(\CMap(), Name$)
      ; Eintrag laden
      *CEntry = \CMap(Name$)
      
      ; Eintrag ist zu alt, bzw. existiert nicht
      If _ENM_Cache_EntryisTooOld(*CEntry) ; Altersgrenze erreicht
        _ENM_Cache_DeleteEntry(*Cache, *CEntry)
        UnlockMutex(\AccessMutex)
      Else
        UnlockMutex(\AccessMutex)
        
        ; Locken und vor löschen sichern
        LockMutex(*CEntry\AcessMutex)
        ; Eintrag wird gelöscht
        If Not *CEntry\deleting
          *CEntry\locked + 1
          *CEntry\AccessCount + 1 ; Zugriffszähler erhöhen
          *CEntry\LastAccess = Date() ; Letzter Zugriff
        Else
          UnlockMutex(*CEntry\AcessMutex)
          ProcedureReturn #Null
        EndIf
        UnlockMutex(*CEntry\AcessMutex)
        
         ; Eintrag ist komprimiert, erst entpacken
        If *CEntry\compressed
          *Mem = zLib_uncompress(*CEntry\Mem, *CEntry\compressed, *CEntry\MemSize)
          If *Mem ; entpacken ok
            FreeMemory(*CEntry\Mem)
            *CEntry\compressed = #False
            *CEntry\Mem = *Mem
          EndIf
        EndIf
        
        ; Entry - Objekt anlegen
        *CIEntry = AllocateMemory(SizeOf(_Cache_Entry_Data))
        *CIEntry\VTable = ?_ENM_Cache_Entry_VTable
        *CIEntry\Entry = *CEntry
        *CIEntry\Cache = *Cache
        
        ; zurückgeben
        ProcedureReturn *CIEntry
      EndIf
    Else ; Sonst nur Mutex freigeben
      LockMutex(\AccessMutex)
    EndIf
    
    ProcedureReturn #Null
  EndWith
EndProcedure

; Eintrag in den Cache hinzufügen
; Name$: Name des Cache Eintrags, kann später damit angesprochen werden
; *Mem: Speicherblock
; MemSize: Speicherblockgröße
; MaxAge: Maximales Alter des Blocks in Sekunden, für Unbegrenzt auf #NULL lassen
; CopyMem: Soll der Speicher kopiert werden?
Procedure _ENM_Cache_Add(*Cache._Cache_Data, Name$, *Mem, MemSize, MaxAge = #Null, CopyMem = #True)
  Protected *tMem, RtVar
  Name$ = LCase(Name$)
  
  ; zu klein zum cachen, mehr Overhead als Nutzen
  If *Cache\MinimumEntrySize And (MemSize < *Cache\MinimumEntrySize)
    ProcedureReturn #Null
  EndIf
  
  If (MemSize< = *Cache\MaxSize) Or (*Cache\MaxSize = #Null) ; Passt das Rein?
    If CopyMem ; Soll der Speicher kopiert werden?
      ; Hier sollte man prüfen ob der Speicher auch wirklich vorhanden ist
      *tMem = AllocateMemory(MemSize)
      If Not *tMem ; Ui, zu wenig Speicher vorhanden...
        ProcedureReturn #False
      Else ; Kopie erzeugen
        CopyMemory(*Mem, *tMem, MemSize)
      EndIf
    Else
      *tMem = *Mem
    EndIf
  
    ; Macht Platz falls nötig
  If _ENM_Cache_ClearOld(*Cache, MemSize)
    
    ; Und ab in die Map
    LockMutex(*Cache\AccessMutex)
    If Not FindMapElement(*Cache\CMap(), Name$)
      *Cache\CurrentSize + MemSize ; Cache wächst
      ; Eintrag anlegen
      AddMapElement(*Cache\CMap(), Name$)
      ; Alles Eintragen
      With *Cache\CMap()
        ; Alles eintragen
        \Mem = *tMem
        \MemSize = MemSize
        If MaxAge > #Null
          \MaxAge = Date() + MaxAge 
        EndIf
        \LastAccess = Date()
        \AcessMutex = CreateMutex()
      EndWith
      ; Adden ok
      RtVar = #True
      
    ElseIf Not CopyMem ; Speicher muss gelöscht werden
      FreeMemory(*tMem)
    EndIf
    UnlockMutex(*Cache\AccessMutex)
    
    ProcedureReturn RtVar
  EndIf
EndIf
EndProcedure

Procedure _ENM_Cache_Entry_GetMem(*CacheE._Cache_Entry_Data)
  ProcedureReturn *CacheE\Entry\Mem
EndProcedure

Procedure _ENM_Cache_Entry_GetMemSize(*CacheE._Cache_Entry_Data)
  ProcedureReturn *CacheE\Entry\MemSize
EndProcedure

Procedure _ENM_Cache_Entry_Release(*CacheE._Cache_Entry_Data)
  With *CacheE
    LockMutex(\Entry\AcessMutex)
    \Entry\locked - 1
    UnlockMutex(\Entry\AcessMutex)
    FreeMemory(*CacheE)
  EndWith
EndProcedure

CompilerIf #ENM_Sys_AsyncFree
  ; Freigabeprocedure für den Async - Manager
  Procedure _ENM_Cache_Entry_Release_Async(*CacheE._Cache_Entry_Data)
    With *CacheE
      If ENM
        JQ_Add(\Cache\ENM\GBCollector, @_ENM_Cache_Entry_Release(), *CacheE)
      Else
        _ENM_Cache_Entry_Release(*CacheE)
      EndIf
    EndWith
  EndProcedure
CompilerEndIf

Procedure _ENM_Cache_Entry_Delete(*CacheE._Cache_Entry_Data)
  With *CacheE
    
    LockMutex(\Entry\AcessMutex)
    ; Locked - Counter zurücksetzen
    \Entry\locked - 1
    ; Eintrag löschen
    If Not \Entry\deleting
      \Entry\deleting = #True ; löschflag setzen
      UnlockMutex(\Entry\AcessMutex)
      
      ; Andere Threads sperren den Thread auch
      If \Entry\locked
        ; Warten bis Eintrag freigegeben wurde
        While \Entry\locked
          Delay(1)
        Wend
      EndIf
      ; Eintrag löschen
      
      LockMutex(\Cache\AccessMutex)
      _ENM_Cache_DeleteEntry(\Cache, \Entry, #False)
      UnlockMutex(\Cache\AccessMutex)
      
    Else
      UnlockMutex(\Entry\AcessMutex)
    EndIf
    
    ; Objekt freigeben
    FreeMemory(*CacheE)
  EndWith
EndProcedure

CompilerIf #ENM_Sys_AsyncFree
  ; Freigabeprocedure für den Async - Manager
  Procedure _ENM_Cache_Entry_Delete_Async(*CacheE._Cache_Entry_Data)
    With *CacheE
      If ENM
        JQ_Add(\Cache\ENM\GBCollector, @_ENM_Cache_Entry_Delete(), *CacheE)
      Else
        _ENM_Cache_Entry_Delete(*CacheE)
      EndIf
    EndWith
  EndProcedure
CompilerEndIf

; VTable:
DataSection
  _ENM_Cache_VTable:
  Data.i @_ENM_Cache_Add()
  Data.i @_ENM_Cache_Get()
  Data.i @_ENM_Cache_Flush()
  CompilerIf #ENM_Sys_AsyncFree
    Data.i @_ENM_Cache_Free_Async()
  CompilerElse
    Data.i @_ENM_Cache_Free()
  CompilerEndIf
  Data.i @_ENM_Cache_GetSize()
  Data.i @_ENM_Cache_CleanUp()
  Data.i @_ENM_Cache_Config()
  
  _ENM_Cache_Entry_VTable:
  Data.i @_ENM_Cache_Entry_GetMem()
  Data.i @_ENM_Cache_Entry_GetMemSize()
  CompilerIf #ENM_Sys_AsyncFree
    Data.i @_ENM_Cache_Entry_Release_Async()
    Data.i @_ENM_Cache_Entry_Delete_Async()
  CompilerElse
    Data.i @_ENM_Cache_Entry_Release()
    Data.i @_ENM_Cache_Entry_Delete()
  CompilerEndIf
EndDataSection


; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 483
; FirstLine = 448
; Folding = ----
; EnableXP
; CPU = 1