﻿; ------------------------------------------------------------------------------------
; Job Queue for Multi - Threading (Lite Edition Single Thread)
; Author: Alexander Aigner
; PB 4.3+ x86/x64 Ascii/(Unicode)
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
; Compilercheck
; ------------------------------------------------------------------------------------

CompilerIf #PB_Compiler_Thread = #False
  CompilerError "Enable Threadsafe"  
CompilerEndIf

CompilerIf (#PB_Compiler_Processor <> #PB_Processor_x86) And (#PB_Compiler_Processor <> #PB_Processor_x64)
  CompilerError "Unsupported Processor"
CompilerEndIf


; Ab wann soll der Thread schlafen
#JQ_SleepDelay = #ENM_JQ_SleepDelay

; ------------------------------------------------------------------------------------
; Prototypes, Structures & Enumerations
; ------------------------------------------------------------------------------------

Prototype JobQueue_Job(*JobData)

Structure _JobQueue_Job
  *Call.JobQueue_Job ; Job - Procedure
  *Para ; Übergabeparameter
EndStructure

Structure _JobQueue_Worker
  Id.i ; ThreadID
  DoCheck.i ; Check - Signal
EndStructure

Structure _JobQueue_Data
  Thread._JobQueue_Worker
  Mutex.i ; Kontroll Mutex
  *Stack.tStack ; Stack
  JobCount.i ; Ausstehende Jobs
  SleepingTimeout.l ; Nach wie viel Zeit soll geschlafen werden?
  StopQueue.a ; Keine weiteren Jobs können mehr hinzugefügt werden
  Sleeping.a ; Thread schläft ;)
  Stop.a ; Deinit - Flag
  Stoped.a ; Deinit - AnswerFlag
EndStructure

Structure _JobQueue_WorkerInit
  *JData._JobQueue_Data
EndStructure

; ------------------------------------------------------------------------------------
; Worker Internal
; ------------------------------------------------------------------------------------

; ist ein Check-Signal da?
Macro _JQ_CheckForCheckSignal()
  If *Worker\DoCheck
    *Worker\DoCheck = #False
    Break 2
  EndIf
EndMacro

; Sendet ein Check-Signal an einen Worker
Macro _JQ_SendCheckSignal(_JobQueue)
  _JobQueue\Thread\DoCheck = #True
EndMacro

; Worker Thread: Arbeitet die Jobs ab
; Der Worker bestimmt wie schnell die gesamte Queue ist, je weniger Overhead desto besser

Procedure JQ_Worker(*Init._JobQueue_WorkerInit)  
  Protected LastJobDone
  Protected *Job._JobQueue_Job
  Protected *JobQueue._JobQueue_Data = *Init\JData
  Protected *Worker._JobQueue_Worker = *JobQueue\Thread
  Protected Mutex = *JobQueue\Mutex
  
  FreeMemory(*Init) ; Brauchen wir nicht mehr
  LastJobDone = ElapsedMilliseconds() ; Letzte Job-Ausführung
  
  With *JobQueue
    Repeat ; Und los gehts
      If \Stop ; JobQueue entfernen
        \Stoped + 1 ; Ein Worker weniger
        ProcedureReturn #True ; beendet sich selbst
      EndIf
      
      If \JobCount ; Ausführen wenn keine Pause ist und Jobs da sind
        LockMutex(Mutex) ; Mutex sperren
        
        While Stack_Count(\Stack) ; Jobs abarbeiten
          _JQ_CheckForCheckSignal() ; Worker soll Flags & seine Messages checken
          *Job = Stack_Pop(\Stack)
          UnlockMutex(Mutex) ; während des Calls muss die Mutex nicht gelockt sein
          
          *Job\Call(*Job\Para) ; Job ausführen
          
          LockMutex(Mutex) ; Mutex wieder sperren
          \JobCount-1
          FreeMemory(*Job) ; Job Memory freigeben
        Wend
        
        ; Letzte Jobausführung
        UnlockMutex(Mutex) ; Mutex freigeben
        
        LastJobDone = ElapsedMilliseconds()
      Else
        ; "Spezialität" : Worker schlafenlegen
        If \SleepingTimeout And ((ElapsedMilliseconds()-LastJobDone) >= \SleepingTimeout)
          LockMutex(Mutex)
          \Sleeping = ((Not Stack_Count(\Stack)) Or #Null)
          UnlockMutex(Mutex)
          If \Sleeping
            ProcedureReturn #Null
          EndIf
        EndIf
        
        Delay(1) ; CPU schonen
      EndIf
      
    ForEver
  EndWith
EndProcedure


Macro _JQ_LaunchThread(_JQ)
  *Thread = AllocateMemory(SizeOf(_JobQueue_WorkerInit))
  *Thread\JData = _JQ
  _JQ#\Thread\Id = CreateThread(@JQ_Worker(), *Thread) ; Worker starten
EndMacro

; ------------------------------------------------------------------------------------
; Public Functions
; ------------------------------------------------------------------------------------

; Initialisiert eine neue Job Queue, Worker = Anzahl Threads
Procedure JQ_Init()
  Protected i, Mutex = CreateMutex(), *JobQueue._JobQueue_Data = AllocateMemory(SizeOf(_JobQueue_Data))
  Protected *Thread._JobQueue_WorkerInit
  LockMutex(Mutex) ; Mutex sperren
  With *JobQueue
    ; Stacks Initialisieren
    \Stack = Stack_New()
    \Mutex = Mutex
    \SleepingTimeout = #JQ_SleepDelay
    _JQ_LaunchThread(*JobQueue)
  EndWith
  UnlockMutex(Mutex) ; Mutex freigeben
  ProcedureReturn *JobQueue
EndProcedure

; Gibt eine Job Queue Frei
Procedure JQ_DeInit(*JobQueue._JobQueue_Data)
  Protected Worker, cTime
  
  With *JobQueue
    
    If Not \Sleeping ; Alles nicht nötig wenn der eine Thread schläft
      \Stop = #True ; Deinit - Flag setzen
      
      _JQ_SendCheckSignal(*JobQueue) ; Worker sollen Flags prüfen
      
      While Not \Stoped
        Delay(1)
        If \Sleeping : Break : EndIf ; Nötig.
      Wend ; Warten bis alle Threads beendet haben
      
    EndIf
    
    While Stack_Count(\Stack)
      FreeMemory(Stack_Pop(\Stack))
    Wend
    Stack_Free(\Stack)
    
    ; Ein Lock-Unlock zyklus am Ende schadet nie.
    LockMutex(\Mutex):UnlockMutex(\Mutex)
    
    ; Alles andere freigeben
    FreeMutex(\Mutex)
    FreeMemory(*JobQueue)
    ProcedureReturn #True
  EndWith
EndProcedure

; Fügt einen Job hinzu (Rückgabewert: JobID für JobQueue_JobFinished()/JobQueue_JobReturnValue())
Procedure JQ_Add(*JobQueue._JobQueue_Data, *JobProcedure.JobQueue_Job, Parameter.i)
  Protected *Job._JobQueue_Job = AllocateMemory(SizeOf(_JobQueue_Job))
  Protected *Thread._JobQueue_WorkerInit
  
  With *JobQueue
    ; Job erstellen
    *Job\Call = *JobProcedure
    *Job\Para = Parameter
    
    LockMutex(\Mutex) ; Mutex sperren
    If \StopQueue ; Queue wurde beendet
      FreeMemory(*Job)
      UnlockMutex(\Mutex)
      ProcedureReturn #False
    EndIf
    If \Sleeping ; Thread schläft...
      \Sleeping = #False
      _JQ_LaunchThread(*JobQueue)
    EndIf
    
    \JobCount + 1
    
    Stack_PushF(\Stack, *Job)
    UnlockMutex(\Mutex); Mutex freigeben
    
    ProcedureReturn #True
  EndWith
EndProcedure

Procedure JQ_Count(*JobQueue._JobQueue_Data)
  With *JobQueue
    ProcedureReturn \JobCount
  EndWith
EndProcedure

; Wartet bis alles abgearbeitet wurde und stopt optional die Queue
Procedure JQ_WaitUntilFinish(*JobQueue._JobQueue_Data, StopQueue = #True)
  With *JobQueue
    LockMutex(\Mutex)
    \StopQueue = StopQueue
    UnlockMutex(\Mutex)
    While *JobQueue\JobCount
      Delay(1)  
    Wend
  EndWith
EndProcedure

; Stopt die Queue, keine Jobs können mehr gecallt werden
Procedure JQ_StopQueue(*JobQueue._JobQueue_Data)
  With *JobQueue
    LockMutex(\Mutex)
    \StopQueue = #True
    UnlockMutex(\Mutex)
  EndWith
EndProcedure

; IDE Options = PureBasic 4.60 Beta 3 (Windows - x64)
; CursorPosition = 244
; FirstLine = 205
; Folding = --
; EnableThread
; EnableXP
; Executable = E:\Dokumente und Einstellungen\Alexander\Desktop\demo.exe
; CPU = 1
; DisableDebugger
; EnableCompileCount = 867
; EnableBuildCount = 1
; EnableExeConstant